/*--------------------------------------------------------------
    trackpeerd - peer tracking agent
    MAC table monitoring module   macmon.c
      Written by  H.Goto, Sep 2005
      Modified by H.Goto, Nov 2005
      http://www.sc.isc.tohoku.ac.jp/~hgot/
--------------------------------------------------------------*/

/*----------------------------------------------------------------------
  Copyright (C) 2005  Hideaki Goto

  All Rights Reserved.

  Permission is hereby granted, free of charge, to any person obtaining
  a copy of this software and associated documentation files (the
  "Software"), to deal in the Software without restriction, including
  without limitation the rights to use, copy, modify, merge, publish,
  distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to
  the following conditions:

  The above copyright notice and this permission notice shall be
  included in all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
  IN NO EVENT SHALL THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR
  ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
  WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

  Except as contained in this notice, the name(s) of the copyright
  holder(s) shall not be used in advertising or otherwise to promote
  the sale, use or other dealings in this Software without prior
  written authorization of the copyright holder(s).
----------------------------------------------------------------------*/

#include	"trackpeerd.h"
#include	"macmon.h"

#define	MACTableSize	1024
#define	MACExpire	60	/* (XXsec + Kernel's ARP table expiration time) */
#define	AliveInterval	180	/* Log Alive_host every 3min. */

#define	ARPDEV	"/proc/net/arp"

		/* MAC Table */
static int	mac_count = 0;
static MACTable_item	MACTable[ MACTableSize ];

static time_t	lastreptime = 0;




/*--------------------------------------------------------------
    MAC table handling
--------------------------------------------------------------*/

static void format_macinfo(char *buf, MACTable_item *macinfo){
	struct in_addr	addr;
	char	eaddrstr_s[20];
	char	addrstr_s[20];
	addr.s_addr = macinfo->ip_src;
	strcpy(eaddrstr_s,ether_ntoa(&macinfo->mac_src));
	strcpy(addrstr_s,inet_ntoa(addr));
	sprintf(buf, "time=%lld ip=%s mac=%s dev=%s", \
		(long long)macinfo->lasttime, \
		addrstr_s, eaddrstr_s, macinfo->devname);
}


static void MAC_update(MACTable_item *macinfo){
	int	i;
	int	pos;
	time_t	currtime;
	char	logstr[80];

	time(&currtime);

	/* lock should be here, if necessary */

	/* find the matching entry */
	pos = -1;
	for ( i=0 ; i<mac_count ; i++ ){
		if ( macinfo->ip_src == MACTable[i].ip_src ){
			pos = i;  break;
		}
	}
	/* insert a new entry */
	if ( pos < 0 ){
		if ( mac_count >= MACTableSize ){
			syslog(SYSLOG_PRIORITY, "Warning: MAC Table overflow.");
		}
		else{	format_macinfo(logstr,macinfo);
			syslog(SYSLOG_PRIORITY, "New_host: %s", logstr);
			MACTable[ mac_count++ ] = *macinfo;
		}
	}
	else{	if ( 0 == memcmp((void *)&macinfo->mac_src, \
			(void *)&MACTable[pos].mac_src, sizeof(struct ether_addr)) ){
			MACTable[pos].lasttime = currtime;	/* update time */
		}
		else{	/* duplicate MAC address detected */
			MACTable[pos].lasttime = currtime;
			format_macinfo(logstr,&MACTable[pos]);
			syslog(SYSLOG_PRIORITY, "Expired_host: %s",logstr);
			format_macinfo(logstr,macinfo);
			syslog(SYSLOG_PRIORITY, "New_host: %s", logstr);
			MACTable[pos] = *macinfo;
		}
	}

	/* unlock should be here, if necessary */
}




/*--------------------------------------------------------------
    MAC table monitoring thread
--------------------------------------------------------------*/

void *macmon_thread(void *pargs){
	FILE	*fp;
	int	i;
	char    prm[6][80];
	char	arpent[80];
	struct ether_addr	*macp;
	MACTable_item mt;
	time_t	currtime;
	char	logstr[80];
	int	loglevel = *(int *)pargs;

	time(&lastreptime);
	openlog("trackpeerd", LOG_PID, SYSLOG_FACILITY);
	if ( strlen(capt_dev) == 0 ){
		syslog(SYSLOG_PRIORITY, "System: Monitoring the ARP table.");
	}
	else{	syslog(SYSLOG_PRIORITY, "System: Monitoring the ARP table for dev=%s.",capt_dev);
	}

	while(1){
		/* expire old entries */
		time(&currtime);
		for ( i=0 ; i<mac_count ; i++ ){
			if ( currtime - MACTable[i].lasttime > MACExpire ){
				MACTable[i].lasttime = currtime;
				format_macinfo(logstr,&MACTable[i]);
				syslog(SYSLOG_PRIORITY, "Expired_host: %s",logstr);
				MACTable[i] = MACTable[mac_count-1];
				mac_count--;  i--;
			}
		}

		if ( NULL == (fp = fopen(ARPDEV,"r")) )  continue;
		while( ! feof(fp) ){
			if ( NULL == fgets(arpent, 80, fp) )  break;
			if ( strncmp(arpent,"IP ",3) == 0 )  continue;
			prm[5][0] = '\0'; 
			sscanf(arpent,"%s %s %s %s %s %s", \
				prm[0], prm[1], prm[2], prm[3], prm[4], prm[5]);
			if ( prm[5][0] == '\0' )  continue;
			if ( strlen(capt_dev) != 0 ){
				if ( strcmp(capt_dev,prm[5]) != 0 )  continue;
			}
			mt.lasttime = currtime;
			if ( 0 >= inet_pton(AF_INET,prm[0],(void *)&mt.ip_src) )  continue;
			if ( NULL == (macp = ether_aton(prm[3])) )  continue;
			memcpy((void *)&mt.mac_src, \
				(void *)macp,sizeof(struct ether_addr));
			mt.devname[15] = '\0';
			strncpy(mt.devname,prm[5],15);
			MAC_update(&mt);

//			printf("%08X\t%s\n",(int)mt.ip_src,ether_ntoa(&mt.mac_src));
		}
		fclose(fp);

		if ( loglevel == 2 ){
			if ( currtime - lastreptime >= AliveInterval ){
				lastreptime = currtime;
				for ( i=0 ; i<mac_count ; i++ ){
					format_macinfo(logstr,&MACTable[i]);
					syslog(SYSLOG_PRIORITY, "Alive_host: %s",logstr);
				}
			}
		}

		sleep(1);
	}
}



